# Widgets.pm - The Widgets Class that provides HTML related widgets.
# Created by James Pattie, 2005-03-08.

# Copyright (c) 2005 Xperience, Inc. http://www.pcxperience.com/
# All rights reserved.  This program is free software; you can redistribute it
# and/or modify it under the same terms as Perl itself.

package HTMLObject::Widgets;
use strict;
use HTMLObject::ErrorBase;
use HTMLObject::Base;
use vars qw($AUTOLOAD $VERSION @ISA @EXPORT @EXPORT_OK);

require Exporter;

@ISA = qw(HTMLObject::ErrorBase Exporter AutoLoader);
@EXPORT = qw(
);

$VERSION = '2.27';

=head1 NAME

HTMLObject::Widgets - HTML Widgets for HTMLObject users.

=head1 SYNOPSIS

  use HTMLObject::Widgets;
  my $widgets = HTMLObject::Widgets->new();

  $doc->print($widgets->generatePicker());  # fill in the parameters to the picker code.

=head1 DESCRIPTION

HTMLObject::Widgets provides a centralized module for all the extra
methods that do non-standard html.

For example, the Date and Color pickers.

=head1 Exported FUNCTIONS

=over 4

=item ref new()

  instantiates the object.

=cut
sub new
{
  my $class = shift;
  my $self = $class->SUPER::new(@_);

  if (!$self->HTMLObject::Widgets::isValid)
  {
    return $self;
  }

  $self->{baseObj} = HTMLObject::Base->new();

  return $self;
}

=item bool isValid()

 returns 1 if everything is ok, 0 otherwise.

=cut
sub isValid
{
  my $self = shift;

  # make sure our Parent class is valid.
  if (!$self->SUPER::isValid())
  {
    $self->prefixError();
    return 0;
  }

  # validate our parameters.

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return 0;
  }
  return 1;
}

=item hash generatePickerCode(type, baseUrl, phrase, width, height, form, itemName, itemValue, windowName, onClick, onChange, link, class, linkClass, year, seperator, displayPrevNextDateLinks)

 Returns the HTML fragment needed to display the phrase as a link
 that allows the user to select a color or date and display their
 selection in the edit field to the right of the link and the
 javascript code to Include the necessary javascript library.

 If an error occured, we return undef and set the error string, so use
 $obj->error() to determine if an error happened and then
 $obj->errorMessage to get the message to display.

 The returned hash has the body and javascriptIncludes entries
 defined so you can just pass the hash to the print() method
 or extract the code yourself.

 For developers needing to get the link/Phrase part seperate
 from the input field definition, use the _link_ and _input_
 entries in the returned hash to get this information.

 requires:
  type: 'color', 'date'
  baseUrl: The url of the server, minus the /htmlobject part, that
    the /htmlobject directory exists at.  Can be empty to indicate
    the current server.
  form: name of the form we are currently in.
  itemName: name of the text field we are creating.
  windowName: name of the window we are going to create.
  phrase:  the string to display.

 optional:
  onClick:  specify any JavaScript code you need executed when the
            link is selected.  This is only valid when link = true.
  onChange: specify any JavaScript code you want executed when the
            edit field is modified. This will only be run if the
            validation code we output returned true to indicate
            you have a valid color or date specified.
  width: defaults to 640
  height: defaults to 550
  itemValue: the value to display in the text field.  defaults to "".
  link: true (1) - display the phrase as a link, false (0) - don't
    make it a link.
  class: specify the CSS class you want the phrase displayed in.
  linkClass: specify the CSS class the link is part of.
  year: only valid when type = 'date'.  Specifies the year to default
        to if the user doesn't specify the year when entering the
        date.
  seperator: only valid when type = 'date'.  Specifies the date
             seperator to work with.  Defaults to '-'.
  displayPrevNextDateLinks: true (1) - display the Prev/Next links,
    false (0) - don't display the Prev/Next links.  Defaults to
    true (1).

  Notes: when displaying a date picker, 2 extra links are now
  output around the text box, if displayPrevNextDateLinks is true.
  They are < and >, to allow you to quickly change the current date 1
  day backwards or forwards.  The output will look like:
  Date < [text box] >, if phrase = "Date".

=cut

sub generatePickerCode
{
  my $self = shift;
  my %args = ( type => "color", width => "640", height => "550", form => "",
               itemName => "", itemValue => "", windowName => "", phrase => "", onChange => "",
               link => 1, year => "", seperator => "-", class => "", linkClass => "", onClick => "",
               baseUrl => "", displayPrevNextDateLinks => 1, @_ );
  my $type = $args{type};
  my $width = $args{width};
  my $height = $args{height};
  my $form = $args{form};
  my $itemName = $args{itemName};
  my $itemValue = $args{itemValue};
  my $windowName = $args{windowName};
  my $phrase = $args{phrase};
  my $onChange = $args{onChange};
  my $link = $args{link};
  my $year = $args{year};
  my $seperator = $args{seperator};
  my $class = $args{class};
  my $linkClass = $args{linkClass};
  my $onClick = $args{onClick};
  my $baseUrl = $args{baseUrl};
  my $displayPrevNextDateLinks = $args{displayPrevNextDateLinks};
  my %result = ();
  my $result = "";

  # fixup the link value
  $link = ($link eq "true" ? 1 : ($link eq "false" ? 0 : $link));

  my $errorStr = "";

  if ($type !~ /^(color|date)$/)
  {
    $self->invalid("type", $type);
  }
  if ($width !~ /^(\d+)$/)
  {
    $self->invalid("width", $width);
  }
  if ($height !~ /^(\d+)$/)
  {
    $self->invalid("height", $height);
  }
  if ($form eq "")
  {
    $self->missing("form");
  }
  if ($itemName eq "")
  {
    $self->missing("itemName");
  }
  if ($windowName eq "")
  {
    $windowName = $type;  # set it equal to the type (date, color)
  }
  if ($phrase eq "")
  {
    $self->missing("phrase");
  }
  if ($link !~ /^(1|0)$/)
  {
    $self->invalid("link", $link, "Can only be 1 or 0.");
  }
  if ($type =~ /^(date)$/)
  {
    if ($year !~ /^(\d{4})$/)
    {
      $self->invalid("year", $year);
    }
    if ($seperator !~ /^(-|\\|\/)$/)
    {
      $self->invalid("seperator", $seperator);
    }
    $displayPrevNextDateLinks = ($displayPrevNextDateLinks eq "true" ? 1 : ($displayPrevNextDateLinks eq "false" ? 0 : $displayPrevNextDateLinks));
  }

  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return undef;
  }

  # formEncode the phrase to protect from xss.
  $phrase = $self->baseObj->formEncode(string => $phrase, sequence => "formatting");

  # build up the form entry.
  my $jsUrl = $baseUrl . "/htmlobject/";

  my $onClickCode = "window." . ($type eq "color" ? "colorField" : "dateField") . "=document.$form.$itemName; " . $type . "Win=window.open('$jsUrl" . ($type eq "color" ? "color_picker.html" : "calendar.html") . "', '$windowName', 'width=$width,height=$height,resizable,scrollbars,status'); " . ($onClick ? $onClick . ($onClick !~ /;$/ ? ";" : "") . " " : "") . "return false;";

  my $linkStr = ($link ? "<a href=\"#\" " . ($linkClass ? "class=\"$linkClass\" " : "") . "onclick=\"$onClickCode\" onmouseover=\"window.status='PopUp " . ($type eq "color" ? "Color" : "Calendar") . " Picker'; return true;\">" : "") . ($class ? "<span class=\"$class\">" : "") . $phrase . ($class ? "</span>" : "") . ($link ? "</a>" : "");

  my $inputStr = ($type eq "date" && $displayPrevNextDateLinks ? qq{<a href="#" } . ($linkClass ? "class=\"$linkClass\" " : "") . qq{ onclick="calcDatePrev(document.$form.$itemName); return false;" onmouseover="window.status='Previous Day'; return true;">&lt;</a> } : "") . "<input type=\"text\" name=\"$itemName\" value=\"$itemValue\"" . (exists $args{disabled} ? " disabled=\"$args{disabled}\"" : "") . " onchange=\"javascript: if (" . ($type eq "color" ? "isValidColor(this)" : "fixupISODate(this, '$seperator', '$year')") . ") { " . ($onChange ? "$onChange" : "") . " }\"" . ($type eq "date" ? " size=\"10\" maxlength=\"10\"" : "") . ">" . ($type eq "date" && $displayPrevNextDateLinks ? qq{ <a href="#" } . ($linkClass ? "class=\"$linkClass\" " : "") . qq{ onclick="calcDateNext(document.$form.$itemName); return false;" onmouseover="window.status='Next Day'; return true;">&gt;</a>} : "");

  $result{body} = $linkStr . " " . $inputStr;
  $result{_link_} = $linkStr;
  $result{_input_} = $inputStr;
  $result{javascriptIncludes} = "<script language=\"JavaScript1.5\" src=\"$jsUrl" . "js/form_methods.js\" type=\"text/javascript\"></script>";

  return %result;
}

=item % generatePopulator(name, rows, required, options, manual, optionsTypes, customLabel, htmlTemplate, selectLocation, clearPhrase,
tableClass, tableStyle, selectClass, selectStyle)

 required:
   name - name of the "populator" widget.  This is the prefix
     value that all generated form items will start with.
   rows - how many rows need to be generated.  Defaults to 1.
   required - how many rows are required to be filled in by the user.
     Defaults to 1.
   options - -Options array as provided to the
     HTMLObject::Form->generate() method.  Used to determine how many
     form items per row to generate.

     Each entry is a hash with the following values:
       label - value to display in the select box for this entry.
       data - array ref of column entries.

     All entries must have the same number of elements in their
     data array.
   manual - (boolean) if true (1), then we just generate the necessary
     javascript since the caller MUST have populated the template,
     profile and data hashes.  if false (0), then we generate the html
     template snippet, profile and data structures and the necessary
     javascript.

 optional:
   optionsTypes - array of hashes that defines what each columns form
     field is to be.  If specified as a string equal to:
       text - all form items are text fields
       datePicker - all form items are date pickers
       colorPicker - all form items are color pickers
     If not specified, it defaults to "text".
     If you provide the array of hashes, the hashes must be valid
     data structures that the HTMLObject::Form->generate() would
     accept as part of its data hash.  You do not need to define the
     name of the form item (as in the data hash you would pass to
     generate()), just the attributes for the form item, like
     -Type, -Label, -Value, etc.
   customLabel - if optionsTypes is one of our special cases:
     text, datePicker, colorPicker
     then you can specify customLabel to provide the label for each
     form item generated, since they are all going to be the same.
     Defaults to "".
   htmlTemplate - user defined html snippet that represents the users
     layout for a row.  If defined and manual = 0, we use it for each
     generated row substituting the field_x_y (where y >= 0 and < the
     number of columns to be generated) values with the form names
     being generated.  The template should only define <td> cells
     and not the parent <tr> cell.
     Ex:
     htmlTemplate = "<td>#LRFI=field_x_0#</td><td>#LRFI=field_x_1#</td>"
     where we will substitute field_x_ with the field name and the
     row number we are currently processing when we generate the html.
   selectLocation - left or right.  specifies if the select box should
     be to the left of the rows or to the right.  Defaults to 'right'.
   clearPhrase - phrase to use for the Clear Row buttons.  Defaults to
     'clear'.
   tableClass - class to apply to the outer table.  Defaults to ''.
   tableStyle - style to apply to the outer table.  Defaults to ''.
   selectClass - class to apply to the select box.  Defaults to ''.
   selectStyle - style to apply to the select box.  Defaults to ''.

 returns:
   hash with the following entries:
     html - generated html hash (contains javascript, body).  The body
       is the html template replacement for the generate() method.
     data - generated data hash that will be used by the generate()
       method to actually create the necessary form items
     profile - generated profile hash to specify what form items are
       required and/or how to validate them if there are special
       validation checks needed.

 summary:

 The generatePopulator() method will output a complex form selection
 widget that uses a select box to determine the values to populate the
 next available row.  The rows will be all the same, but there is no
 requirement that the columns have to be all the same.  You can specify
 a very complex input screen by defining the optionsTypes array.

 The generated html will consist of a table that takes up 100% of the
 available space.  It will be split into 2 columns, with the select box
 in the left or right column, based upon the selectLocation parameter.
 The other column will have a child table defined and for each generated
 row of form items, we wrap the htmlTemplate in the <tr></tr> tags.
 If htmlTemplate is not specified, then each form item is defined as:
   <td align="right" valign="top">#RL=field_x_y#</td>
   <td align="left">#FI=field_x_y#</td>

 where field_x_y means:
   field = name
   x = row #, starting at 0
   y = column #, starting at 0

 The generated javascript will store the options in a javascript array
 of objects with values = col0, col1, col2, etc. allowing quick and
 easy lookup and assignment.  The select box, will only have the label
 to be displayed for each selection.  The selectedIndex value will be
 used to index into the javascript array to determine the data to
 populate.  This way we don't have to do string splitting, etc., which
 is expensive and can have issues if the user wants to use our seperator
 value in their data.

 The javascript will attempt to find an empty row and populate it.  If
 no empty rows are found, then it will popup an alert and inform the
 user of this condition.

 On submission, if required > 0, then the javascript will first attempt
 to make sure that at least required rows were populated and then will
 make sure that required rows starting at index 0 and incrementing by
 1 are populated to meet the FormValidator requirements as specified
 by the profile.


 Sample usage for specifying in the HTMLObject::Form data hash:

  data{foo}->{bar} = { -Type => "populator", -Rows => 4, -Required => 1,
    -Options => \%selectOptions, manual => 0,
    -OptionsTypes => "datePicker", -SelectLocation => "left" };

=cut
sub generatePopulator
{
  my $self = shift;
  my %args = ( name => "", rows => 1, required => 1, options => [], manual => 0,
               optionsTypes => "text", htmlTemplate => "", selectLocation => "right",
               clearPhrase => "clear", tableClass => "", tableStyle => "",
               selectClass => "", selectStyle => "", customLabel => "", @_ );

  my $name = $args{name};
  my $rows = $args{rows};
  my $required = $args{required};
  my $options = $args{options};
  my $manual = $args{manual};
  my $optionsTypes = $args{optionsTypes};
  my $customLabel = $args{customLabel};
  my $htmlTemplate = $args{htmlTemplate};
  my $selectLocation = $args{selectLocation};
  my $clearPhrase = $args{clearPhrase};
  my $tableClass = $args{tableClass};
  my $tableStyle = $args{tableStyle};
  my $selectClass = $args{selectClass};
  my $selectStyle = $args{selectStyle};
  my $numColumns = -1;

  my %result = ( html => { javascript => "", body => "" }, data => {}, profile => {} );

  # fixup the manual value
  $manual = ($manual eq "true" ? 1 : ($manual eq "false" ? 0 : $manual));

  # do validation of input.
  if ($name eq "")
  {
    $self->missing("name");
  }
  if ($rows !~ /^(\d+)$/)
  {
    $self->invalid("rows", $rows, "must be a positive number >= 0");
  }
  if ($required !~ /^(\d+)$/)
  {
    $self->invalid("required", $required, "must be a positive number >= 0");
  }
  if (ref ($options) ne "ARRAY")
  {
    $self->invalid("options", ref($options), "must be an array ref");
  }
  elsif (scalar @{$options} == 0)
  {
    $self->missing("options", "must have at least 1 entry");
  }
  else
  {
    # now walk over the entries and make sure they appear valid.
    for (my $i=0; $i < scalar @{$options}; $i++)
    {
      if (ref ($options->[$i]) ne "HASH")
      {
        $self->invalid("options[$i]", ref($options->[$i]), "must be a hash ref");
      }
      else
      {
        if (! exists $options->[$i]->{label})
        {
          $self->missing("options[$i]->{label}");
        }
        elsif (length $options->[$i]->{label} == 0)
        {
          $self->invalid("options[$i]->{label}", "", "You must specify the label to display");
        }
        if (! exists $options->[$i]->{data})
        {
          $self->missing("options[$i]->{data}");
        }
        elsif (ref ($options->[$i]->{data}) ne "ARRAY")
        {
          $self->invalid("options[$i]->{data}", ref($options->[$i]->{data}), "must be an array ref");
        }
        else
        {
          my $tmpColumns = scalar @{$options->[$i]->{data}};
          $numColumns = $tmpColumns  if ($i == 0);
          if ($tmpColumns == 0 || $tmpColumns != $numColumns)
          {
            $self->invalid("options[$i]->{data}", $tmpColumns, "length can not be = 0 or is not = '$numColumns'");
          }
        }
      }
    }
  }
  if ($manual !~ /^(0|1)$/)
  {
    $self->invalid("manual", $manual, "must be 0 or 1");
  }
  if (defined $optionsTypes)
  {
    if (ref ($optionsTypes) ne "ARRAY")
    {
      if ($optionsTypes !~ /^(text|datePicker|colorPicker)$/)
      {
        $self->invalid("optionsTypes", $optionsTypes, "valid values: text, datePicker, colorPicker or array of data hash entries");
      }
    }
    else
    {
      # now loop over the array and make sure we have hashes that look semi valid.
      for (my $i=0; $i < scalar @{$optionsTypes}; $i++)
      {
        if (ref($optionsTypes->[$i]) ne "HASH")
        {
          $self->invalid("optionsTypes[$i]", ref($optionsTypes->[$i]), "must be an array ref");
        }
        else
        {
          # check for -Type as a minimum.
          if (!exists $optionsTypes->[$i]->{-Type})
          {
            $self->missing("optionsTypes[$i]->{-Type}", "-Type must be defined");
          }
        }
      }
    }
  }
  if ($selectLocation !~ /^(left|right)$/)
  {
    $self->invalid("selectLocation", $selectLocation, "must be 'left' or 'right'");
  }
  if ($self->numInvalid() > 0 || $self->numMissing() > 0)
  {
    $self->error($self->genErrorString("all"));
    return %result;
  }

  # at this point, we appear to be valid.

  # lets start generating the javascript data structures.
  my $jscript = "\n";

  my $jscriptData = $name . "PopulatorData";
  $jscript .= "// data for populator widget = '$name'\n";
  $jscript .= "var $jscriptData = new Array;\n";
  for (my $i=0; $i < scalar @{$options}; $i++)
  {
    $jscript .= $jscriptData . "[$i] = new Array(";
    for (my $j=0; $j < scalar @{$options->[$i]->{data}}; $j++)
    {
      (my $value = $options->[$i]->{data}->[$j]) =~ s/'/\\'/g;
      $jscript .= ", " if ($j > 0);
      $jscript .= "'$value'";
    }
    $jscript .= ");\n";
  }

  my $jscriptSelect = $name . "PopulatorSelect";
  my $jscriptClearRow = $name . "PopulatorClearRow";
  my $jscriptMoveRow = $name . "PopulatorMoveRow";
  my $jscriptConsolidateRows = $name . "PopulatorConsolidateRows";
  my $jscriptEnforceRequiredRows = $name . "PopulatorEnforceRequiredRows";
  # now generate the actual populate method for when the user selects an entry from the select box.
  $jscript .= qq{
function $jscriptSelect(field)
{
  var index = field.selectedIndex;

  if (index == -1 || index > field.options.length)
  {
    alert("selected index = '" + index + "' is invalid!\\nYou must select an item to populate.");
    return;
  }

  // find an empty row.
  var rowFound = false;
  for (var i=0; i < $rows && !rowFound; i++)
  {
    var rowEmpty = true;
    for (var j=0; j < $numColumns; j++)
    {
      var t = "result = field.form.$name" + "_" + i + "_" + j + ".value;";
      var result = "";
      try {
        eval(t.toString());
      }
      catch (e)
      {
        alert(e);
        return;
      }
      if (result.length > 0)
      {
        rowEmpty = false;
      }
    }
    if (rowEmpty)
    {
      rowFound = true;
      // do the population from the selectedIndex value.
      for (var j=0; j < $numColumns; j++)
      {
        var t = "field.form.$name" + "_" + i + "_" + j + ".value = $jscriptData" + "[" + index + "][" + j + "];";
        try {
          eval(t.toString());
        }
        catch (e)
        {
          alert(e);
          return;
        }
      }
    }
  }
  if (!rowFound)
  {
    alert("You do not have any empty rows to be populated!\\nPlease empty a row and try again.");
    return;
  }
}

function $jscriptClearRow(index)
{
  if (index < 0 || index > $rows)
  {
    alert("$jscriptClearRow(): index = '" + index + "' is out of bounds!");
    return;
  }

  for (var j=0; j < $numColumns; j++)
  {
    var t = "field.form.$name" + "_" + index + "_" + j + ".value = '';";
    try {
      eval(t.toString());
    }
    catch (e)
    {
      alert(e);
      return;
    }
  }
}

function $jscriptMoveRow(srcRow, destRow)
{
  if (srcRow < 0 || srcRow > $rows)
  {
    alert("$jscriptMoveRow(): srcRow = '" + srcRow + "' is out of bounds!");
    return;
  }

  if (dstRow < 0 || destRow > $rows)
  {
    alert("$jscriptMoveRow(): destRow = '" + destRow + "' is out of bounds!");
    return;
  }

  for (var j=0; j < $numColumns; j++)
  {
    var t = "field.form.$name" + "_" + destRow + "_" + j + ".value = field.form.$name" + "_" + srcRow + "_" + j + ".value;";
    try {
      eval(t.toString());
    }
    catch (e)
    {
      alert(e);
      return;
    }
  }

  // now clear the srcRow
  $jscriptClearRow(srcRow);
}

function $jscriptConsolidateRows(field)
{
  for (var i=0; i < $rows; i++)
  {
    var rowEmpty = true;
    for (var j=0; j < $numColumns; j++)
    {
      var t = "result = field.form.$name" + "_" + i + "_" + j + ".value;";
      var result = "";
      try {
        eval(t.toString());
      }
      catch (e)
      {
        alert(e);
        return;
      }
      if (result.length > 0)
      {
        rowEmpty = false;
      }
    }
    if (rowEmpty)
    {
      var rowFound = false;
      // look for a non-empty row that is above this row (increasing numerically).
      for (var k=i+1; k < $rows && !rowFound; k++)
      {
        var nextRowEmpty = true;
        for (var j=0; j < $numColumns; j++)
        {
          var t = "result = field.form.$name" + "_" + k + "_" + j + ".value;";
          var result = "";
          try {
            eval(t.toString());
          }
          catch (e)
          {
            alert(e);
            return;
          }
          if (result.length > 0)
          {
            nextRowEmpty = false;
          }
        }
        if (!nextRowEmpty)
        {
          // we found a non-empty row, so move it.
          $jscriptMoveRow(k, i);
          rowFound = true;
        }
      }
      if (!rowFound)
      {
        return;  // we are done, since there are no more empty rows to consolidate up.
      }
    }
  }
}

function $jscriptEnforceRequiredRows(field)
{
  // first we need to consolidate the rows, just to make life easier.
  $jscriptConsolidateRows(field);

  // now check to make sure the number of required rows have been met.
  if ($required > 0)
  {
    var numFilledRows = 0;
    for (var i=0; i < $rows; i++)
    {
      // check and see if this row is full.
      var rowEmpty = true;
      for (var j=0; j < $numColumns; j++)
      {
        var t = "result = field.form.$name" + "_" + i + "_" + j + ".value;";
        var result = "";
        try {
          eval(t.toString());
        }
        catch (e)
        {
          alert(e);
          return;
        }
        if (result.length > 0)
        {
          rowEmpty = false;
        }
      }
      if (!rowEmpty)
      {
        numFilledRows++;
      }
    }
    if (numFilledRows < $required)
    {
      alert("You only have '" + numFilledRows + "' rows filled!\\n$required are required to be filled.");
      return false; // abort the submit.
    }
  }

  return true;  // go ahead and do the submit.
}

};

  # update the result hash.
  $result{html}->{javascript} = $jscript;
  $result{html}->{onsubmit} = qq{if (!$jscriptEnforceRequiredRows(this.$name} . "Select) { return false; }";

  # now start building up the html, data and profile outputs.
  if (!$manual)
  {
    # build up the html template snippet.
    my $css = ($tableClass ? qq{class="$tableClass" } : "") . ($tableStyle ? qq{style="$tableStyle"} : "");

    # build up the left and right cell contents.
    my $leftCellContent = "";
    my $rightCellContent = "";
    my $selectTemplate = qq{#F=$name} . "Select#";
    my $rowsTemplate = qq{<table $css border="0" width="100%" cellpadding="2" cellspacing="0">\n};
    for (my $i=0; $i < $rows; $i++)
    {
      $rowsTemplate .= "        <tr>\n";
      my $tempTemplate = $htmlTemplate;
      for (my $j=0; $j < $numColumns; $j++)
      {
        my $fname = $name . "_$i" . "_$j";
        if ($tempTemplate)
        {
          $tempTemplate =~ s/field_x_$j/$fname/g;
        }
        else
        {
          $rowsTemplate .= qq{          <td align="right" valign="top">#RL=$fname#</td>\n};
          $rowsTemplate .= qq{          <td align="left">#FI=$fname#</td>\n};
        }
      }
      if ($tempTemplate)
      {
        $tempTemplate =~ s/^(\s*)(.+)$/          $2/;
        $rowsTemplate .= $tempTemplate . ($tempTemplate !~ /\n$/ ? "\n" : "");
      }
      $rowsTemplate .= "        </tr>\n";
    }
    $rowsTemplate .= "      </table>";

    if ($selectLocation eq "left")
    {
      $leftCellContent = $selectTemplate;
      $rightCellContent = $rowsTemplate;
    }
    else
    {
      $leftCellContent = $rowsTemplate;
      $rightCellContent = $selectTemplate;
    }

    my $body = qq{
<table $css border="0" width="100%" cellpadding="2" cellspacing="0">
  <tr>
    <td align="center">
      $leftCellContent
    </td>
    <td align="center">
      $rightCellContent
    </td>
  </tr>
</table>
};
    $result{html}->{body} = $body;

    # build up the data hash, first the select box.
    my %selectOptions = ();
    for (my $i=0; $i < scalar @{$options}; $i++)
    {
      push @{$selectOptions{names}}, $options->[$i]->{label};
      push @{$selectOptions{values}}, $i; # the value is not actually used by the widget.
    }

    my %css = ();
    $css{class} = $selectClass if ($selectClass);
    $css{style} = $selectStyle if ($selectStyle);

    $result{data}->{$name . "Select"} = { -Type => "select", -Value => "", -Label => "", -Options => \%selectOptions, %css };

    # now generate the rows.
    for (my $i=0; $i < $rows; $i++)
    {
      for (my $j=0; $j < $numColumns; $j++)
      {
        my $fname = $name . "_$i" . "_$j";
        $result{data}->{$fname} = {};
        if (ref ($optionsTypes) eq "ARRAY")
        {
          # clone the optionsTypes entry for this column.
          foreach my $entry (keys %{$optionsTypes->[$j]})
          {
            $result{data}->{$fname}->{$entry} = $optionsTypes->[$j]->{$entry};
          }
        }
        else
        {
          $result{data}->{$fname}->{-Type} = $optionsTypes;
          $result{data}->{$fname}->{-Value} = "";
          $result{data}->{$fname}->{-Label} = $customLabel;
        }
      }
    }

    # build up the profile structure.
    for (my $i=0; $i < $required; $i++)
    {
      for (my $j=0; $j < $numColumns; $j++)
      {
        my $fname = $name . "_$i" . "_$j";
        push @{$result{profile}->{required}}, $fname;

        # handle any special checks here
      }
    }
  }

  return %result;
}

=back

=cut

1;
__END__

=head1 AUTHOR

James A. Pattie, htmlobject@pcxperience.com

=head1 SEE ALSO

perl(1), HTMLObject::Base(3), HTMLObject::FrameSet(3), HTMLObject::ReadCookie(3), HTMLObject::Form(3), HTMLObject::ErrorBase(3).

=cut
